// This file is part of eMule 
//
///LANCAST
//
// Written by moosetea, enjoy
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// This allows lan clients to add each other as peers, it could be
// very buggy and needs alot more work.
//
// Thanks to codeproject for help with the multicast code, 
// which works but is a bit of a shambles.

// Base includes needed by this class (and every class)
#include "StdAfx.h"
#include "emule.h"

// We also need the class defintion, and specific includes
#include "LanCast.h"
#include "packets.h"
#include "ED2KLink.h"

// CLanCast

CLanCast::CLanCast(CPreferences* in_prefs)
{

	prefs = in_prefs;

	bStarted = false;
	if (GetLancastEnabled())
	{
		udp_timer = SetTimer(0,4329,10000,CLanCast::UDPTimerProc);
		Start();
	}
}

bool CLanCast::SendPacket(Packet* packet)
{

		char* sendbuffer = new char[packet->size+2];

		// Standard UDP Header
		memcpy(sendbuffer,packet->GetUDPHeader(),2);

		// The packet Data
		memcpy(sendbuffer+2,packet->pBuffer,packet->size);

		// Send the LanCast packet
		if(m_SendSocket.SendTo(sendbuffer, packet->size+10, (SOCKADDR*)&m_saHostGroup, sizeof(SOCKADDR), 0) == SOCKET_ERROR)
		{
			delete[] sendbuffer;
			return FALSE;
		}
		else
		{
			delete[] sendbuffer;
			return TRUE;
		}

}

void CLanCast::BroadcastHash(CKnownFile* cur_file){


	if (GetLancastEnabled())
		Start();
	else
		Stop();


	if (GetLancastEnabled()){

		CMemFile* file = new CMemFile(100);

		uint32 dwID = GetLancastIP();
		uint16 nPort = GetEmulePort();

		file->Write(&dwID,4);
		file->Write(&nPort,2);
		file->Write(cur_file->GetFileHash(),16);

		Packet* packet = new Packet(file,OP_LANCASTPROT);
		packet->opcode = OP_HASH;

		SendPacket(packet);
		delete packet;
		delete file;
	
	}
}

void CLanCast::RecieveHash(char* pachPacket, uint32 nSize){

	CSafeMemFile* offeredfiles = new CSafeMemFile((BYTE*)pachPacket,nSize);

	uint32 dwID = 0;
	uint16 nPort = 0; 
	uchar filehash[16];

	// Read Hash Data
	offeredfiles->Read(&dwID,4);
	offeredfiles->Read(&nPort,2);
	offeredfiles->Read(&filehash,16);

	 // Debug Info for Devs - The following lines should be removed in release
	char buffer[50];
	buffer[0] = 0;
	for (uint16 i = 0;i != 16;i++)
		sprintf(buffer,"%s%02X",buffer,filehash[i]);
	AddDebugLogLine(false,"Lancast : Just Recieved %s",buffer);

	if (CPartFile* file = theApp.downloadqueue->GetFileByID((uchar*)filehash)) {

		// Debug Info for Devs - the following lines can be removed in release
		AddDebugLogLine(false,"Lancast : Adding source for %s",buffer);

		// Create a new Updown client for this source (with no server info), userid is the LAN ip address
		CUpDownClient* newsource = new CUpDownClient(nPort,dwID,0,0,file);

		// Add this LAN uldl client to the download queue, this also will check for duplicate sockets and "merge" them
		theApp.downloadqueue->CheckAndAddSource(file,newsource);

	}

	delete offeredfiles;

}


CLanCast::~CLanCast()
{
	Stop();
}

void CLanCast::OnReceive(int nErrorCode)
{

	char buffer[5000];
	CString addressbuffer;
	uint32 port;

	int32 length = ReceiveFrom(buffer,5000,addressbuffer,port);

	// if lancast is enabled proccess
	if (GetLancastEnabled()){

		// If this isnt a packet from ourselves
		if (inet_addr(addressbuffer) != GetLancastIP())
		{

			if ((uint8)buffer[0] == OP_LANCASTPROT && length != SOCKET_ERROR)
				ProcessPacket(buffer+2,length-2,buffer[1]);

		}

	}

}

bool CLanCast::ProcessPacket(char* packet, uint32 size, uint8 opcode){

	try{

		switch(opcode){
			case OP_HASH :{	
				
				RecieveHash(packet,size);
				break;
			}
			default:
				AddLogLine(false,"Lancast : Unknown OpCode");
				return false;
		}

		return true;
	}
	catch(...){
		AddLogLine(false,"Lancast : Error receiving multicast data");
		return false;
	}
}

void CALLBACK CLanCast::UDPTimerProc(HWND hwnd, UINT uMsg,UINT_PTR idEvent,DWORD dwTime){
		theApp.sharedfiles->NextLANBroadcast();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLanCast Private member functions - MultiCast Network Code

BOOL CLanCast::CreateReceivingSocket(LPCTSTR strGroupIP, UINT nGroupPort)
{
	/* Create socket for receiving packets from multicast group */
	if(!Create(nGroupPort, SOCK_DGRAM, FD_READ))
		return FALSE;

	BOOL bMultipleApps = TRUE;		/* allow reuse of local port if needed */
	SetSockOpt(SO_REUSEADDR, (void*)&bMultipleApps, sizeof(BOOL), SOL_SOCKET);

	/* Fill m_saHostGroup_in for sending datagrams */
	memset(&m_saHostGroup, 0, sizeof(m_saHostGroup));
	m_saHostGroup.sin_family = AF_INET;
	m_saHostGroup.sin_addr.s_addr = inet_addr(strGroupIP);
	m_saHostGroup.sin_port = htons((USHORT)nGroupPort);

	/* Join the multicast group */
	m_mrMReq.imr_multiaddr.s_addr = inet_addr(strGroupIP);	/* group addr */ 
	m_mrMReq.imr_interface.s_addr = htons(INADDR_ANY);		/* use default */ 
	if(setsockopt(m_hSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char FAR *)&m_mrMReq, sizeof(m_mrMReq)) < 0)
		return FALSE;

	return TRUE;
}

BOOL CLanCast::CreateSendingSocket(UINT nTTL, BOOL bLoopBack)
{
	if(!m_SendSocket.Create(0, SOCK_DGRAM, 0))		// Create an unconnected UDP socket
		return FALSE;

	if(!SetTTL(nTTL))								// Set Time to Live as specified by user
		AddLogLine(false,"Lancast : Warning, Error Setting Time to live");

	SetLoopBack(bLoopBack);

	return TRUE;
}

void CLanCast::Start()
{
	if (!bStarted){

		char t_strLanCastIP[17];
		sprintf(t_strLanCastIP,"%i.%i.%i.%i",(uint8)GetLancastIP(),(uint8)(GetLancastIP()>>8),(uint8)(GetLancastIP()>>16),(uint8)(GetLancastIP()>>24));

		// Try and join the multicast group - 224.0.0.1, using the prefs port and no loopback
		if(!JoinGroup ("224.0.0.1", GetLancastPort() , 50,false))
			AddLogLine(false,"LanCast : Emule Lan Id has been set to %u (%s), Failed to Join LAN Broadcast Group ( 224.0.0.1 ) on port %i",GetLancastIP(),t_strLanCastIP,GetLancastPort());
		else {
			AddLogLine(false,"LanCast : Emule Lan Id has been set to %u (%s), Sucessfully Joined LAN Broadcast Group ( 224.0.0.1 ) on port %i",GetLancastIP(),t_strLanCastIP,GetLancastPort());
		}
	}

}

void CLanCast::Stop()
{
	if (bStarted){
		if(!LeaveGroup())
			if (theApp.emuledlg)	//sometimes method is called in app shutdown [TwoBottle Mod]
				AddLogLine(false,"Lancast : Failed to Leave LAN Broadcast Group ( 224.0.0.1 ) on port %i",GetLancastPort());
		else
			if (theApp.emuledlg)	//sometimes method is called in app shutdown [TwoBottle Mod]
				AddLogLine(false,"Lancast : Sucessfully Left LAN Broadcast Group ( 224.0.0.1 ) on port %i",GetLancastPort());
	}
}


BOOL CLanCast::JoinGroup(CString GroupIP, UINT nGroupPort, UINT nTTL, BOOL bLoopback)
{

	bStarted = false;

	if(!CreateReceivingSocket(GroupIP, nGroupPort))		/* Create Socket for receiving and join the host group */
		return FALSE;
	if(!CreateSendingSocket(nTTL, bLoopback))			/* Create Socket for sending */
		return FALSE;

	bStarted = true;

	return TRUE;
}

BOOL CLanCast::LeaveGroup()
{

	bStarted = false;

	if(setsockopt (m_hSocket, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char FAR *)&m_mrMReq, sizeof(m_mrMReq)) < 0)
		return FALSE;

	m_SendSocket.Close();		// Close sending socket
	Close();					// Close receving socket
	return TRUE;
}

BOOL CLanCast::SetTTL(UINT nTTL)
{
	/* Set Time to Live to parameter TTL */
	if(m_SendSocket.SetSockOpt(IP_MULTICAST_TTL, &nTTL, sizeof(int), IPPROTO_IP) == 0)
		return FALSE;		/* Error Setting TTL */
	else
		return TRUE;		/* else TTL set successfully */
}

void CLanCast::SetLoopBack(BOOL bLoop)
{
	/* Set LOOPBACK option to TRUE OR FALSE according to IsLoop parameter */
	int nLoopBack = (int)bLoop;

	// Try to manually set the loopback, ie tell dozer to ignlore packets from itself (if setloopback is false)
	m_SendSocket.SetSockOpt(IP_MULTICAST_LOOP, &nLoopBack, sizeof(int), IPPROTO_IP);

}